open CursorAPI

(********************************************************************
  module TableCursor
    an instance of cursor API implemented by the Slit data structure
 ********************************************************************)

module Table :
  sig
    type t
    val entry_num  : t -> int
    val field_num  : t -> int
    val field_name : int -> t -> string
    val get        : int -> int -> t -> string
    val set        : string -> int -> int -> t -> t
    val insert     : int -> t -> t
    val remove     : int -> t -> t
    val load_csv   : in_channel -> t
    val save_csv   : out_channel -> t -> unit
  end 
   =
  struct
    type t = string array array
    let entry_num    tbl = Array.length tbl - 1
    let field_num    tbl = Array.length (tbl.(0))
    let field_name i tbl = tbl.(0).(i)
    let get      e f tbl = tbl.(e+1).(f)
    let set    s e f tbl = tbl.(e+1).(f) <- s; tbl
    let insert e tbl = let e = e+1 in
                       let newtbl = Array.make (Array.length tbl + 1) (Array.make (field_num tbl) "") in
                         Array.blit tbl 0 newtbl    0  e;
                         Array.blit tbl e newtbl (e+1) (Array.length tbl - e);
                         newtbl
    let remove e tbl = let e = e+1 in
                       let newtbl = Array.make (Array.length tbl - 1) tbl.(0) in
                         Array.blit tbl    0  newtbl 0 e;
                         Array.blit tbl (e+1) newtbl e (Array.length tbl - (e+1));
                         newtbl

    let load_csv ch =
      let getline () = Array.of_list (Str.split (Str.regexp " *, *") (input_line ch)) in
      let fieldnames = getline () in
      let fl  = Array.length fieldnames in
      let lst = ref [fieldnames]  in
        (try  while true do
           let entry = getline () in
             if fl = Array.length entry then
               lst := entry :: !lst done
         with End_of_file -> ());
         Array.of_list (List.rev !lst)

    let save_csv ch tbl =
      let fend = field_num tbl - 1 in
        Array.iter (Array.iteri (fun f a ->
          output_string ch a;
          output_char   ch (if f = fend then '\n' else ',');
        )) tbl
  end



type cursor = LV1 of Table.t * int (* 0 or 1 *)
            | LV2 of Table.t * int
            | LV3 of Table.t * int * int
            | LV4 of Table.t * int * int * int (* 0 or 1 *)

let prev = function
  | LV1(t,1)              -> LV1(t,0)
  | LV2(t,i)   when i > 0 -> LV2(t,i-1)
  | LV3(t,i,j) when j > 0 -> LV3(t,i,j-1)
  | LV4(t,i,j,1)          -> LV4(t,i,j,0)

let next = function
  | LV1(t,0)                              -> LV1(t,1)
  | LV2(t,i)   when i < Table.entry_num t -> LV2(t,i+1)
  | LV3(t,i,j) when j < Table.field_num t -> LV3(t,i,j+1)
  | LV4(t,i,j,0)                          -> LV4(t,i,j,1)

let up = function
  | LV2(t,i)     -> LV1(t,0)
  | LV3(t,i,j)   -> LV2(t,i)
  | LV4(t,i,j,k) -> LV3(t,i,j)

let down = function
  | LV1(t,0)                              -> LV2(t,0)
  | LV2(t,i)   when i < Table.entry_num t -> LV3(t,i,0)
  | LV3(t,i,j) when j < Table.field_num t -> LV4(t,i,j,0)

let get = function
  | LV1(t,0)                                -> CursorAPI.Node( "Root",  [] )
  | LV2(t,i)     when i < Table.entry_num t -> CursorAPI.Node( "Entry", [] )
  | LV3(t,i,j)   when j < Table.field_num t -> CursorAPI.Node( Table.field_name j t, [] )
  | LV4(t,i,j,0)                            -> CursorAPI.Data( Table.get i j t )

let set (CursorAPI.Data s) = function
  | LV4(t,i,j,0) -> LV4(Table.set s i j t,i,j,0)

let insertA (CursorAPI.Node ("Entry",[])) = function
  | LV2(t,i) -> LV2(Table.insert i t, i)

let insertB (CursorAPI.Node ("Entry",[])) = function
  | LV2(t,i) -> LV2(Table.insert i t, i+1)

let remove = function
  | LV2(t,i) when i < Table.entry_num t -> LV2(Table.remove i t, i)

let input_table_cursor ch =
  LV1(Table.load_csv ch,0)

let output_table_cursor ch = function
  | LV1(t,_)
  | LV2(t,_)
  | LV3(t,_,_)
  | LV4(t,_,_,_) -> Table.save_csv ch t
